import builtins
import gettext
import json
import os
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import override

from archinstall.lib.command import SysCommand
from archinstall.lib.exceptions import SysCallError
from archinstall.lib.log import debug
from archinstall.lib.utils.util import running_from_iso


@dataclass
class Language:
	abbr: str
	name_en: str
	translation: gettext.NullTranslations
	translation_percent: int
	translated_lang: str | None
	console_font: str | None = None

	@property
	def display_name(self) -> str:
		name = self.name_en
		return f'{name} ({self.translation_percent}%)'

	def is_match(self, lang_or_translated_lang: str) -> bool:
		if self.name_en == lang_or_translated_lang:
			return True
		elif self.translated_lang == lang_or_translated_lang:
			return True
		return False

	def json(self) -> str:
		return self.name_en


_DEFAULT_FONT = 'default8x16'
_ENV_FONT = os.environ.get('FONT')


class TranslationHandler:
	def __init__(self) -> None:
		self._base_pot = 'base.pot'
		self._languages = 'languages.json'
		self._active_language: Language | None = None
		self._font_backup: Path | None = None
		self._cmap_backup: Path | None = None
		self._using_env_font: bool = False

		self._total_messages = self._get_total_active_messages()
		self._translated_languages = self._get_translations()

	@property
	def translated_languages(self) -> list[Language]:
		return self._translated_languages

	@property
	def active_font(self) -> str | None:
		if self._active_language is not None:
			return self._active_language.console_font
		return None

	def _set_font(self, font_name: str | None) -> bool:
		"""Set the console font via setfont. Only runs on ISO. Returns True on success."""
		if not running_from_iso():
			return False

		target = font_name or _DEFAULT_FONT
		try:
			SysCommand(['setfont', target])
			return True
		except SysCallError as err:
			debug(f'Failed to set console font {target}: {err}')
			return False

	def save_console_font(self) -> None:
		"""Save the current console font (with unicode map) and console map to temp files."""
		if not running_from_iso():
			return

		try:
			font_fd, font_path = tempfile.mkstemp(prefix='archinstall_font_')
			cmap_fd, cmap_path = tempfile.mkstemp(prefix='archinstall_cmap_')
			os.close(font_fd)
			os.close(cmap_fd)
			self._font_backup = Path(font_path)
			self._cmap_backup = Path(cmap_path)
			SysCommand(['setfont', '-O', str(self._font_backup), '-om', str(self._cmap_backup)])
		except SysCallError as err:
			debug(f'Failed to save console font: {err}')
			self._font_backup = None
			self._cmap_backup = None

	def restore_console_font(self) -> None:
		"""Restore console font (with unicode map) and console map from backup."""
		if not running_from_iso():
			return

		if self._font_backup is None or not self._font_backup.exists():
			return

		cmd = ['setfont', str(self._font_backup)]
		if self._cmap_backup is not None and self._cmap_backup.exists():
			cmd += ['-m', str(self._cmap_backup)]
		try:
			SysCommand(cmd)
		except SysCallError as err:
			debug(f'Failed to restore console font: {err}')

		self._font_backup.unlink(missing_ok=True)
		self._font_backup = None
		if self._cmap_backup is not None:
			self._cmap_backup.unlink(missing_ok=True)
			self._cmap_backup = None

	def _get_translations(self) -> list[Language]:
		"""
		Load all translated languages and return a list of such
		"""
		mappings = self._load_language_mappings()
		defined_languages = self._provided_translations()

		languages = []

		for short_form in defined_languages:
			mapping_entry: dict[str, str] = next(filter(lambda x: x['abbr'] == short_form, mappings))
			abbr = mapping_entry['abbr']
			lang = mapping_entry['lang']
			translated_lang = mapping_entry.get('translated_lang', None)
			console_font = mapping_entry.get('console_font', None)

			try:
				# get a translation for a specific language
				translation = gettext.translation('base', localedir=self._get_locales_dir(), languages=(abbr, lang))

				# calculate the percentage of total translated text to total number of messages
				if abbr == 'en':
					percent = 100
				else:
					num_translations = self._get_catalog_size(translation)
					percent = int((num_translations / self._total_messages) * 100)
					# prevent cases where the .pot file is out of date and the percentage is above 100
					percent = min(100, percent)

				language = Language(abbr, lang, translation, percent, translated_lang, console_font)
				languages.append(language)
			except FileNotFoundError as err:
				raise FileNotFoundError(f"Could not locate language file for '{lang}': {err}")

		return languages

	def _load_language_mappings(self) -> list[dict[str, str]]:
		"""
		Load the mapping table of all known languages
		"""
		locales_dir = self._get_locales_dir()
		languages = Path.joinpath(locales_dir, self._languages)

		with open(languages) as fp:
			return json.load(fp)

	def _get_catalog_size(self, translation: gettext.NullTranslations) -> int:
		"""
		Get the number of translated messages for a translations
		"""
		# this is a very naughty way of retrieving the data but
		# there's no alternative method exposed unfortunately
		catalog = translation._catalog  # type: ignore[attr-defined]
		messages = {k: v for k, v in catalog.items() if k and v}
		return len(messages)

	def _get_total_active_messages(self) -> int:
		"""
		Get total messages that could be translated
		"""
		locales = self._get_locales_dir()
		with open(f'{locales}/{self._base_pot}') as fp:
			lines = fp.readlines()
			msgid_lines = [line for line in lines if 'msgid' in line]

		return len(msgid_lines) - 1  # don't count the first line which contains the metadata

	def get_language_by_name(self, name: str) -> Language:
		"""
		Get a language object by it's name, e.g. English
		"""
		try:
			return next(filter(lambda x: x.name_en == name, self._translated_languages))
		except Exception:
			raise ValueError(f'No language with name found: {name}')

	def get_language_by_abbr(self, abbr: str) -> Language:
		"""
		Get a language object by its abbreviation, e.g. en
		"""
		try:
			return next(filter(lambda x: x.abbr == abbr, self._translated_languages))
		except Exception:
			raise ValueError(f'No language with abbreviation "{abbr}" found')

	def activate(self, language: Language, set_font: bool = True) -> None:
		"""
		Set the provided language as the current translation
		"""
		# The install() call has the side effect of assigning GNUTranslations.gettext to builtins._
		language.translation.install()
		self._active_language = language

		if set_font and not self._using_env_font:
			self._set_font(language.console_font)

	def apply_console_font(self) -> None:
		"""Apply console font from FONT env var or active language mapping.

		If FONT env var is set and valid, use it and skip language mapping.
		If FONT is set but invalid, fall back to language font.
		If FONT is not set, use active language font.
		"""
		if not running_from_iso():
			return

		if _ENV_FONT:
			if self._set_font(_ENV_FONT):
				self._using_env_font = True
				debug(f'Console font set from FONT env var: {_ENV_FONT}')
			else:
				debug(f'FONT={_ENV_FONT} could not be set, falling back to language font mapping')
				if self.active_font:
					self._set_font(self.active_font)
					debug(f'Console font set from language mapping: {self.active_font}')
		elif self.active_font:
			self._set_font(self.active_font)
			debug(f'Console font set from language mapping: {self.active_font}')

	def _get_locales_dir(self) -> Path:
		"""
		Get the locales directory path
		"""
		cur_path = Path(__file__).parent.parent
		locales_dir = Path.joinpath(cur_path, 'locales')
		return locales_dir

	def _provided_translations(self) -> list[str]:
		"""
		Get a list of all known languages
		"""
		locales_dir = self._get_locales_dir()
		filenames = os.listdir(locales_dir)

		translation_files = []
		for filename in filenames:
			if len(filename) == 2 or filename in ['pt_BR', 'zh-CN', 'zh-TW']:
				translation_files.append(filename)

		return translation_files


class _DeferredTranslation:
	def __init__(self, message: str):
		self.message = message

	@override
	def __str__(self) -> str:
		if builtins._ is _DeferredTranslation:  # type: ignore[attr-defined]
			return self.message

		# builtins._ is changed from _DeferredTranslation to GNUTranslations.gettext after
		# Language.activate() is called
		return builtins._(self.message)  # type: ignore[attr-defined]


def tr(message: str) -> str:
	return str(_DeferredTranslation(message))


builtins._ = _DeferredTranslation  # type: ignore[attr-defined]


translation_handler = TranslationHandler()



from collections.abc import Callable
from dataclasses import asdict, is_dataclass
from typing import TYPE_CHECKING, Any

from archinstall.lib.utils.encoding import unicode_ljust, unicode_rjust
from archinstall.tui.rich import BaseRichTable

if TYPE_CHECKING:
	from _typeshed import DataclassInstance


def as_key_value_pair(
	entries: dict[str, str | list[str] | bool],
	ignore_empty: bool = True,
) -> str:
	"""
	Formats key-values as a Rich Table:
		key1	: value1
		key2	: value2
	...
	"""
	table = BaseRichTable()
	table.add_column('key', style='bold', no_wrap=True)
	table.add_column('value', style='white', max_width=70)

	for label, value in entries.items():
		if ignore_empty and not value:
			continue

		if isinstance(value, bool):
			value = 'Yes' if value else 'No'

		if isinstance(value, list):
			value = '\n  '.join(str(val) for val in value)

		table.add_row(label.title(), f': {value}')

	return table.stringify()


def as_columns(entries: list[str], cols: int) -> str:
	"""
	Will format a list into a given number of columns
	"""
	chunks: list[list[str]] = []
	output = ''

	for i in range(0, len(entries), cols):
		chunks.append(entries[i : i + cols])

	for row in chunks:
		out_fmt = '{: <30} ' * len(row)
		output += out_fmt.format(*row) + '\n'

	return output


def _get_values(
	o: DataclassInstance,
	class_formatter: str | Callable | None = None,  # type: ignore[type-arg]  # pyright: ignore[reportMissingTypeArgument]
	filter_list: list[str] = [],
) -> dict[str, Any]:
	"""
	the original values returned a dataclass as dict thru the call to some specific methods
	this version allows thru the parameter class_formatter to call a dynamically selected formatting method.
	Can transmit a filter list to the class_formatter,
	"""
	if class_formatter:
		# if invoked per reference it has to be a standard function or a classmethod.
		# A method of an instance does not make sense
		if callable(class_formatter):
			return class_formatter(o, filter_list)
		# if is invoked by name we restrict it to a method of the class. No need to mess more
		elif hasattr(o, class_formatter) and callable(getattr(o, class_formatter)):
			func = getattr(o, class_formatter)
			return func(filter_list)

		raise ValueError('Unsupported formatting call')
	elif hasattr(o, 'table_data'):
		return o.table_data()
	elif hasattr(o, 'json'):
		return o.json()
	elif is_dataclass(o):
		return asdict(o)
	else:
		return o.__dict__  # type: ignore[unreachable]


def as_table(
	obj: list[Any],
	class_formatter: str | Callable | None = None,  # type: ignore[type-arg]
	filter_list: list[str] = [],
	capitalize: bool = False,
) -> str:
	"""variant of as_table (subtly different code) which has two additional parameters
	filter which is a list of fields which will be shown
	class_formatter a special method to format the outgoing data

	A general comment, the format selected for the output (a string where every data record is separated by newline)
	is for compatibility with a print statement
	As_table_filter can be a drop in replacement for as_table
	"""
	raw_data = [_get_values(o, class_formatter, filter_list) for o in obj]

	# determine the maximum column size
	column_width: dict[str, int] = {}
	for o in raw_data:
		for k, v in o.items():
			if not filter_list or k in filter_list:
				column_width.setdefault(k, 0)
				column_width[k] = max([column_width[k], len(str(v)), len(k)])

	if not filter_list:
		filter_list = list(column_width.keys())

	# create the header lines
	output = ''
	key_list = []
	for key in filter_list:
		width = column_width[key]
		key = key.replace('!', '').replace('_', ' ')

		if capitalize:
			key = key.capitalize()

		key_list.append(unicode_ljust(key, width))

	output += ' | '.join(key_list) + '\n'
	output += '-' * len(output) + '\n'

	# create the data lines
	for record in raw_data:
		obj_data = []
		for key in filter_list:
			width = column_width.get(key, len(key))
			value = record.get(key, '')

			if '!' in key:
				value = '*' * len(value)

			if isinstance(value, (int, float)) or (isinstance(value, str) and value.isnumeric()):
				obj_data.append(unicode_rjust(str(value), width))
			else:
				obj_data.append(unicode_ljust(str(value), width))

		output += ' | '.join(obj_data) + '\n'

	return output



from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path

from archinstall.lib.hardware import SysInfo
from archinstall.lib.models.bootloader import Bootloader, BootloaderConfiguration
from archinstall.lib.models.device import DiskLayoutConfiguration


class BootloaderValidationFailureKind(Enum):
	LimineNonFatBoot = auto()
	LimineLayout = auto()
	BootloaderRequiresUefi = auto()
	EfistubNonFatBoot = auto()


@dataclass(frozen=True)
class BootloaderValidationFailure:
	kind: BootloaderValidationFailureKind
	description: str


def validate_bootloader_layout(
	bootloader_config: BootloaderConfiguration | None,
	disk_config: DiskLayoutConfiguration | None,
) -> BootloaderValidationFailure | None:
	"""Validate bootloader configuration against disk layout.

	Returns a failure with a human-readable description if the configuration
	would produce an unbootable system, or None if it is valid.
	"""
	if not (bootloader_config and disk_config):
		return None

	bootloader = bootloader_config.bootloader

	if bootloader == Bootloader.NO_BOOTLOADER:
		return None

	if bootloader.is_uefi_only() and not SysInfo.has_uefi():
		return BootloaderValidationFailure(
			kind=BootloaderValidationFailureKind.BootloaderRequiresUefi,
			description=f'{bootloader.value} requires a UEFI system.',
		)

	boot_part = next(
		(p for m in disk_config.device_modifications if (p := m.get_boot_partition())),
		None,
	)

	if bootloader == Bootloader.Efistub:
		# The UEFI firmware reads the kernel directly from the boot partition,
		# which must be FAT.
		if boot_part and (boot_part.fs_type is None or not boot_part.fs_type.is_fat()):
			return BootloaderValidationFailure(
				kind=BootloaderValidationFailureKind.EfistubNonFatBoot,
				description='Efistub does not support booting with a non-FAT boot partition.',
			)

	if bootloader == Bootloader.Limine:
		# Limine reads its config and kernels from the boot partition, which
		# must be FAT.
		if boot_part and (boot_part.fs_type is None or not boot_part.fs_type.is_fat()):
			return BootloaderValidationFailure(
				kind=BootloaderValidationFailureKind.LimineNonFatBoot,
				description='Limine does not support booting with a non-FAT boot partition.',
			)

		# When the ESP is the boot partition but mounted outside /boot and
		# UKI is disabled, kernels end up on the root filesystem which
		# Limine cannot access.
		if not bootloader_config.uki:
			efi_part = next(
				(p for m in disk_config.device_modifications if (p := m.get_efi_partition())),
				None,
			)
			if efi_part and efi_part == boot_part and efi_part.mountpoint != Path('/boot'):
				return BootloaderValidationFailure(
					kind=BootloaderValidationFailureKind.LimineLayout,
					description=(
						f'Limine requires kernels on a FAT partition. The ESP is mounted at {efi_part.mountpoint}, '
						'enable UKI or add a separate /boot partition to install Limine.'
					),
				)

	return None



import shutil
from pathlib import Path
from typing import TYPE_CHECKING, override

from archinstall.default_profiles.profile import DisplayServerType, GreeterType, Profile, ProfileType

if TYPE_CHECKING:
	from archinstall.lib.installer import Installer
	from archinstall.lib.models.users import User


_TERMINAL = 'alacritty'
_ASSETS_DIR = Path(__file__).parent / 'niri_dms_assets'


class NiriDmsProfile(Profile):
	def __init__(self) -> None:
		super().__init__(
			'niri - DankMaterialShell',
			ProfileType.WindowMgr,
			support_gfx_driver=True,
			display_server=DisplayServerType.Wayland,
		)

	@property
	@override
	def packages(self) -> list[str]:
		return [
			'niri',
			'dms-shell-niri',
			'polkit',
			'xdg-desktop-portal-gnome',
			'xorg-xwayland',
			'matugen',
			'cava',
			'kimageformats',
			'cups-pk-helper',
			'tuned-ppd',
			_TERMINAL,
		]

	@property
	@override
	def default_greeter_type(self) -> GreeterType:
		return GreeterType.GreetdDms

	@override
	def provision(self, install_session: Installer, users: list[User]) -> None:
		binds = (_ASSETS_DIR / 'dms/binds.kdl').read_text().replace('{{TERMINAL_COMMAND}}', _TERMINAL)

		for user in users:
			home = install_session.target / 'home' / user.username
			niri_dir = home / '.config/niri'
			dms_dir = niri_dir / 'dms'
			dms_dir.mkdir(parents=True, exist_ok=True)

			shutil.copy(_ASSETS_DIR / 'niri.kdl', niri_dir / 'config.kdl')
			for name in ('colors.kdl', 'layout.kdl', 'alttab.kdl', 'outputs.kdl', 'cursor.kdl'):
				shutil.copy(_ASSETS_DIR / 'dms' / name, dms_dir / name)
			(dms_dir / 'binds.kdl').write_text(binds)

			niri_unit_dropin = home / '.config/systemd/user/niri.service.d'
			niri_unit_dropin.mkdir(parents=True, exist_ok=True)
			(niri_unit_dropin / 'dms.conf').write_text('[Unit]\nWants=dms.service\n')

			install_session.arch_chroot(f'chown -R {user.username}:{user.username} /home/{user.username}/.config')



from io import StringIO

from rich import box
from rich.console import Console
from rich.table import Table as RichTable


class BaseRichTable(RichTable):
	def __init__(self) -> None:
		super().__init__(
			box=box.SIMPLE,
			show_header=False,
			pad_edge=False,
			show_edge=False,
		)

	def stringify(self) -> str:
		string_io = StringIO()
		buf = Console(file=string_io, highlight=False)
		buf.print(self)

		_ = string_io.seek(0)
		output = string_io.read()

		return output



from typing import override

from archinstall.default_profiles.desktops.utils import select_seat_access
from archinstall.default_profiles.profile import CustomSetting, DisplayServerType, GreeterType, Profile, ProfileType


class NiriProfile(Profile):
	def __init__(self) -> None:
		super().__init__(
			'niri',
			ProfileType.WindowMgr,
			support_gfx_driver=True,
			display_server=DisplayServerType.Wayland,
		)

		self.custom_settings = {CustomSetting.SeatAccess: None}

	@property
	@override
	def packages(self) -> list[str]:
		additional = []
		if seat := self.custom_settings.get(CustomSetting.SeatAccess, None):
			additional = [seat]

		return [
			'niri',
			'alacritty',
			'fuzzel',
			'mako',
			'xorg-xwayland',
			'waybar',
			'swaybg',
			'swayidle',
			'swaylock',
			'xdg-desktop-portal-gnome',
		] + additional

	@property
	@override
	def default_greeter_type(self) -> GreeterType:
		return GreeterType.Lightdm

	@property
	@override
	def services(self) -> list[str]:
		if pref := self.custom_settings.get(CustomSetting.SeatAccess, None):
			return [pref]
		return []

	@override
	async def do_on_select(self) -> None:
		default = self.custom_settings.get(CustomSetting.SeatAccess, None)
		seat_access = await select_seat_access(self.name, default)
		self.custom_settings[CustomSetting.SeatAccess] = seat_access.value



from dataclasses import dataclass
from enum import StrEnum, auto
from typing import Any, NotRequired, Self, TypedDict, override

from archinstall.lib.models.config import SubConfig
from archinstall.lib.translationhandler import tr


class PowerManagement(StrEnum):
	POWER_PROFILES_DAEMON = 'power-profiles-daemon'
	TUNED = auto()


class PowerManagementConfigSerialization(TypedDict):
	power_management: str


class BluetoothConfigSerialization(TypedDict):
	enabled: bool


class Audio(StrEnum):
	NO_AUDIO = 'No audio server'
	PIPEWIRE = auto()
	PULSEAUDIO = auto()


class AudioConfigSerialization(TypedDict):
	audio: str


class PrintServiceConfigSerialization(TypedDict):
	enabled: bool


class Firewall(StrEnum):
	UFW = auto()
	FWD = 'firewalld'


class FirewallConfigSerialization(TypedDict):
	firewall: str


class FontPackage(StrEnum):
	NOTO = 'noto-fonts'
	EMOJI = 'noto-fonts-emoji'
	CJK = 'noto-fonts-cjk'
	LIBERATION = 'ttf-liberation'
	DEJAVU = 'ttf-dejavu'

	def description(self) -> str:
		match self:
			case FontPackage.NOTO:
				return tr('Unicode font coverage for most languages')
			case FontPackage.EMOJI:
				return tr('color emoji for browsers and apps')
			case FontPackage.CJK:
				return tr('Chinese, Japanese, Korean characters')
			case FontPackage.LIBERATION:
				return tr('Arial/Times/Courier replacement, Cyrillic support for Steam/games')
			case FontPackage.DEJAVU:
				return tr('wide Unicode coverage, good fallback font')


class FontsConfigSerialization(TypedDict):
	fonts: list[str]


class ZramAlgorithm(StrEnum):
	ZSTD = auto()
	LZO_RLE = 'lzo-rle'
	LZO = auto()
	LZ4 = auto()
	LZ4HC = auto()


class ApplicationSerialization(TypedDict):
	bluetooth_config: NotRequired[BluetoothConfigSerialization]
	audio_config: NotRequired[AudioConfigSerialization]
	power_management_config: NotRequired[PowerManagementConfigSerialization]
	print_service_config: NotRequired[PrintServiceConfigSerialization]
	firewall_config: NotRequired[FirewallConfigSerialization]
	fonts_config: NotRequired[FontsConfigSerialization]


@dataclass
class AudioConfiguration:
	audio: Audio

	def json(self) -> AudioConfigSerialization:
		return {
			'audio': self.audio.value,
		}

	@classmethod
	def parse_arg(cls, arg: dict[str, Any]) -> Self:
		return cls(
			Audio(arg['audio']),
		)


@dataclass
class BluetoothConfiguration:
	enabled: bool

	def json(self) -> BluetoothConfigSerialization:
		return {'enabled': self.enabled}

	@classmethod
	def parse_arg(cls, arg: BluetoothConfigSerialization) -> Self:
		return cls(arg['enabled'])


@dataclass
class PowerManagementConfiguration:
	power_management: PowerManagement

	def json(self) -> PowerManagementConfigSerialization:
		return {
			'power_management': self.power_management.value,
		}

	@classmethod
	def parse_arg(cls, arg: PowerManagementConfigSerialization) -> Self:
		return cls(
			PowerManagement(arg['power_management']),
		)


@dataclass
class PrintServiceConfiguration:
	enabled: bool

	def json(self) -> PrintServiceConfigSerialization:
		return {'enabled': self.enabled}

	@classmethod
	def parse_arg(cls, arg: PrintServiceConfigSerialization) -> Self:
		return cls(arg['enabled'])


@dataclass
class FirewallConfiguration:
	firewall: Firewall

	def json(self) -> FirewallConfigSerialization:
		return {
			'firewall': self.firewall.value,
		}

	@classmethod
	def parse_arg(cls, arg: dict[str, Any]) -> Self:
		return cls(
			Firewall(arg['firewall']),
		)


@dataclass
class FontsConfiguration:
	fonts: list[FontPackage]

	def json(self) -> FontsConfigSerialization:
		return {'fonts': [f.value for f in self.fonts]}

	@classmethod
	def parse_arg(cls, arg: FontsConfigSerialization) -> Self:
		return cls(fonts=[FontPackage(f) for f in arg['fonts']])


@dataclass(frozen=True)
class ZramConfiguration(SubConfig):
	enabled: bool
	algorithm: ZramAlgorithm = ZramAlgorithm.ZSTD

	@classmethod
	def parse_arg(cls, arg: bool | dict[str, Any]) -> Self:
		if isinstance(arg, bool):
			return cls(enabled=arg)

		enabled = arg.get('enabled', True)
		algo = arg.get('algorithm', arg.get('algo', ZramAlgorithm.ZSTD.value))
		return cls(enabled=enabled, algorithm=ZramAlgorithm(algo))

	@override
	def json(self) -> dict[str, bool | str]:
		return {
			'enabled': self.enabled,
			'algorithm': self.algorithm.value,
		}

	@override
	def summary(self) -> list[str] | None:
		out: list[str] = []

		if self.enabled:
			out.append(tr('Zram enabled'))

			out.append(tr('Zram algorithm {}').format(self.algorithm))
			return out

		return None


@dataclass
class ApplicationConfiguration(SubConfig):
	bluetooth_config: BluetoothConfiguration | None = None
	audio_config: AudioConfiguration | None = None
	power_management_config: PowerManagementConfiguration | None = None
	print_service_config: PrintServiceConfiguration | None = None
	firewall_config: FirewallConfiguration | None = None
	fonts_config: FontsConfiguration | None = None

	@classmethod
	def parse_arg(
		cls,
		args: dict[str, Any] | None = None,
		old_audio_config: dict[str, Any] | None = None,
	) -> Self:
		app_config = cls()

		if args and (bluetooth_config := args.get('bluetooth_config')) is not None:
			app_config.bluetooth_config = BluetoothConfiguration.parse_arg(bluetooth_config)

		# deprecated: backwards compatibility
		if old_audio_config is not None:
			app_config.audio_config = AudioConfiguration.parse_arg(old_audio_config)

		if args and (audio_config := args.get('audio_config')) is not None:
			app_config.audio_config = AudioConfiguration.parse_arg(audio_config)

		if args and (power_management_config := args.get('power_management_config')) is not None:
			app_config.power_management_config = PowerManagementConfiguration.parse_arg(power_management_config)

		if args and (print_service_config := args.get('print_service_config')) is not None:
			app_config.print_service_config = PrintServiceConfiguration.parse_arg(print_service_config)

		if args and (firewall_config := args.get('firewall_config')) is not None:
			app_config.firewall_config = FirewallConfiguration.parse_arg(firewall_config)

		if args and (fonts_config := args.get('fonts_config')) is not None:
			app_config.fonts_config = FontsConfiguration.parse_arg(fonts_config)

		return app_config

	@override
	def json(self) -> ApplicationSerialization:
		config: ApplicationSerialization = {}

		if self.bluetooth_config:
			config['bluetooth_config'] = self.bluetooth_config.json()

		if self.audio_config:
			config['audio_config'] = self.audio_config.json()

		if self.power_management_config:
			config['power_management_config'] = self.power_management_config.json()

		if self.print_service_config:
			config['print_service_config'] = self.print_service_config.json()

		if self.firewall_config:
			config['firewall_config'] = self.firewall_config.json()

		if self.fonts_config:
			config['fonts_config'] = self.fonts_config.json()

		return config

	@override
	def summary(self) -> list[str]:
		out: list[str] = []

		if self.bluetooth_config and self.bluetooth_config.enabled:
			out.append(tr('Bluetooth enabled'))

		if self.audio_config:
			out.append(tr('Audio server "{}"').format(self.audio_config.audio))

		if self.power_management_config:
			out.append(tr('Power management "{}"').format(self.power_management_config.power_management))

		if self.print_service_config and self.print_service_config.enabled:
			out.append(tr('Print service enabled'))

		if self.firewall_config:
			out.append(tr('Firewall "{}"').format(self.firewall_config.firewall))

		if self.fonts_config and self.fonts_config.fonts:
			fonts = ', '.join(f.value for f in self.fonts_config.fonts)
			out.append(tr('Extra fonts "{}"').format(fonts))

		return out



from enum import StrEnum
from typing import override

from archinstall.default_profiles.profile import CustomSetting, DisplayServerType, GreeterType, Profile, ProfileType
from archinstall.lib.menu.helpers import Selection
from archinstall.lib.packages.packages import available_package, package_group_info
from archinstall.lib.translationhandler import tr
from archinstall.tui.menu_item import MenuItem, MenuItemGroup
from archinstall.tui.result import ResultType


class PlasmaFlavor(StrEnum):
	Meta = 'plasma-meta'
	Plasma = 'plasma'
	Desktop = 'plasma-desktop'

	def show(self) -> str:
		match self:
			case PlasmaFlavor.Meta:
				return f'{self.value} ({tr("Recommended")})'
			case PlasmaFlavor.Plasma | PlasmaFlavor.Desktop:
				return self.value

	def package_details(self) -> str:
		ty = ''
		details = ''
		desc = ''

		match self:
			case PlasmaFlavor.Meta:
				ty = tr('Package')
				desc = tr('Curated selection of KDE Plasma packages')
				info = available_package(self.value)

				if info is not None:
					details = tr('Dependencies') + '\n'
					details += '\n'.join(f'- {entry}' for entry in info.get_depends_on)
			case PlasmaFlavor.Plasma:
				ty = tr('Package group')
				desc = tr('Extensive KDE Plasma installation')
				group = package_group_info(self.value)

				if group is not None:
					details = tr('Packages in group') + '\n'
					details += '\n'.join(f'- {entry}' for entry in group.packages)
			case PlasmaFlavor.Desktop:
				ty = tr('Package group')
				desc = tr('Minimal KDE Plasma installation')
				info = available_package(self.value)

				if info is not None:
					details = tr('Dependencies') + '\n'
					details += '\n'.join(f'- {entry}' for entry in info.get_depends_on)

		return f'{tr("Type")}: {ty}\n{tr("Description")}: {desc}\n\n{details}'

	def packages(self) -> list[str]:
		match self:
			case PlasmaFlavor.Meta:
				return ['plasma-meta']
			case PlasmaFlavor.Plasma:
				return ['plasma']
			case PlasmaFlavor.Desktop:
				return ['plasma-desktop']


class PlasmaProfile(Profile):
	def __init__(self) -> None:
		super().__init__(
			'KDE Plasma',
			ProfileType.DesktopEnv,
			support_gfx_driver=True,
			display_server=DisplayServerType.Wayland,
		)

	@property
	@override
	def packages(self) -> list[str]:
		flavor_str = self.custom_settings.get(CustomSetting.PlasmaFlavor)

		if flavor_str is not None:
			flavor = PlasmaFlavor(flavor_str)
			return flavor.packages()
		else:
			return PlasmaFlavor.Meta.packages()  # use plasma-meta as the recommended default

	@property
	@override
	def default_greeter_type(self) -> GreeterType:
		return GreeterType.PlasmaLoginManager

	async def _select_flavor(self) -> None:
		header = tr('Select a flavor of KDE Plasma to install') + '\n'

		items = [
			MenuItem(
				s.show(),
				value=s,
				preview_action=lambda x: x.value.package_details() if x.value else None,
			)
			for s in PlasmaFlavor
		]
		group = MenuItemGroup(items, sort_items=False)

		default = self.custom_settings.get(CustomSetting.PlasmaFlavor, None)
		group.set_default_by_value(default)

		result = await Selection[PlasmaFlavor](
			group,
			header=header,
			allow_skip=False,
			preview_location='right',
		).show()

		if result.type_ == ResultType.Selection:
			self.custom_settings[CustomSetting.PlasmaFlavor] = result.get_value().value

	@override
	async def do_on_select(self) -> None:
		await self._select_flavor()



import re
from pathlib import Path

from archinstall.lib.models.packages import Repository
from archinstall.lib.models.pacman import PacmanConfiguration
from archinstall.lib.pathnames import PACMAN_CONF


class PacmanConfig:
	def __init__(self, target: Path | None):
		self._config_remote_path: Path | None = None

		if target:
			self._config_remote_path = target / PACMAN_CONF.relative_to_root()

		self._repositories: list[Repository] = []

	def enable(self, repo: Repository | list[Repository]) -> None:
		if not isinstance(repo, list):
			repo = [repo]

		self._repositories += repo

	def apply(self) -> None:
		if not self._repositories:
			return

		repos_to_enable = []
		for repo in self._repositories:
			if repo == Repository.Testing:
				repos_to_enable.extend(['core-testing', 'extra-testing', 'multilib-testing'])
			else:
				repos_to_enable.append(repo.value)

		content = PACMAN_CONF.read_text().splitlines(keepends=True)

		for row, line in enumerate(content):
			# Check if this is a commented repository section that needs to be enabled
			match = re.match(r'^#\s*\[(.*)\]', line)

			if match and match.group(1) in repos_to_enable:
				# uncomment the repository section line, properly removing # and any spaces
				content[row] = re.sub(r'^#\s*', '', line)

				# also uncomment the next line (Include statement) if it exists and is commented
				if row + 1 < len(content) and content[row + 1].lstrip().startswith('#'):
					content[row + 1] = re.sub(r'^#\s*', '', content[row + 1])

		# Write the modified content back to the file
		with PACMAN_CONF.open('w') as f:
			f.writelines(content)

	def persist(self) -> None:
		if self._config_remote_path:
			PACMAN_CONF.copy(self._config_remote_path, preserve_metadata=True)

	def configure(self, pacman_config: PacmanConfiguration) -> None:
		"""Apply PacmanConfiguration (Color, ParallelDownloads) to the target system's pacman.conf."""
		if not self._config_remote_path or not self._config_remote_path.exists():
			return

		content = self._config_remote_path.read_text().splitlines()
		result = []

		for line in content:
			if re.match(r'^#?\s*ParallelDownloads', line):
				result.append(f'ParallelDownloads = {pacman_config.parallel_downloads}')
			elif re.match(r'^#?\s*Color\s*$', line):
				result.append('Color' if pacman_config.color else '#Color')
			else:
				result.append(line)

		self._config_remote_path.write_text('\n'.join(result) + '\n')



from abc import ABC, abstractmethod
from typing import Any


class SubConfig(ABC):
	@abstractmethod
	def json(self) -> Any:
		pass

	@abstractmethod
	def summary(self) -> str | list[str] | None:
		pass



from typing import override

from archinstall.lib.menu.abstract_menu import AbstractSubMenu
from archinstall.lib.menu.helpers import Confirmation, Input
from archinstall.lib.models.pacman import PacmanConfiguration
from archinstall.lib.pathnames import PACMAN_CONF
from archinstall.lib.translationhandler import tr
from archinstall.tui.menu_item import MenuItem, MenuItemGroup
from archinstall.tui.result import ResultType


class PacmanMenu(AbstractSubMenu[PacmanConfiguration]):
	def __init__(
		self,
		pacman_conf: PacmanConfiguration,
		advanced: bool = False,
	):
		self._pacman_conf = pacman_conf
		self._advanced = advanced
		menu_options = self._define_menu_options()

		self._item_group = MenuItemGroup(menu_options, sort_items=False, checkmarks=True)
		super().__init__(
			self._item_group,
			config=self._pacman_conf,
			allow_reset=True,
		)

	def _define_menu_options(self) -> list[MenuItem]:
		return [
			MenuItem(
				text=tr('Parallel Downloads'),
				action=select_parallel_downloads,
				value=self._pacman_conf.parallel_downloads,
				preview_action=lambda item: str(item.get_value()),
				key='parallel_downloads',
				enabled=self._advanced,
			),
			MenuItem(
				text=tr('Color'),
				action=select_color,
				value=self._pacman_conf.color,
				preview_action=lambda item: str(item.get_value()),
				key='color',
			),
		]

	@override
	async def show(self) -> PacmanConfiguration | None:
		config = await super().show()

		if config is None:
			return PacmanConfiguration.default()

		_apply_to_live(config.parallel_downloads)

		return config


def _apply_to_live(parallel_downloads: int) -> None:
	"""Apply ParallelDownloads to live system pacman.conf for faster installation."""
	with PACMAN_CONF.open() as f:
		pacman_conf = f.read().split('\n')

	with PACMAN_CONF.open('w') as fwrite:
		for line in pacman_conf:
			if 'ParallelDownloads' in line:
				fwrite.write(f'ParallelDownloads = {parallel_downloads}\n')
			else:
				fwrite.write(f'{line}\n')


async def select_parallel_downloads(preset: int = 5) -> int | None:
	max_recommended = 10

	header = tr('Enter the number of parallel downloads (1-{})').format(max_recommended)

	def validator(s: str) -> str | None:
		try:
			value = int(s)
			if 1 <= value <= max_recommended:
				return None
			return tr('Value must be between 1 and {}').format(max_recommended)
		except Exception:
			return tr('Please enter a valid number')

	result = await Input(
		header=header,
		allow_skip=True,
		allow_reset=True,
		validator_callback=validator,
		default_value=str(preset),
	).show()

	match result.type_:
		case ResultType.Skip:
			return preset
		case ResultType.Reset:
			return 5
		case ResultType.Selection:
			return int(result.get_value())


async def select_color(preset: bool = True) -> bool | None:
	result = await Confirmation(
		header=tr('Enable colored output for pacman'),
		preset=preset,
		allow_skip=True,
	).show()

	match result.type_:
		case ResultType.Skip:
			return preset
		case ResultType.Reset:
			return True
		case ResultType.Selection:
			return result.get_value()



import time
from collections.abc import Iterator
from pathlib import Path
from types import TracebackType
from typing import ClassVar, Self

from archinstall.lib.command import SysCommand, SysCommandWorker
from archinstall.lib.exceptions import SysCallError
from archinstall.lib.log import error


class Boot:
	_active_boot: ClassVar[Self | None] = None

	def __init__(self, path: Path | str):
		if isinstance(path, Path):
			path = str(path)

		self.path = path
		self.container_name = 'archinstall'
		self.session: SysCommandWorker | None = None
		self.ready = False

	def __enter__(self) -> Self:
		if Boot._active_boot and Boot._active_boot.path != self.path:
			raise KeyError('Archinstall only supports booting up one instance and another session is already active.')

		if Boot._active_boot:
			self.session = Boot._active_boot.session
			self.ready = Boot._active_boot.ready
		else:
			# '-P' or --console=pipe  could help us not having to do a bunch
			# of os.write() calls, but instead use pipes (stdin, stdout and stderr) as usual.
			self.session = SysCommandWorker(
				[
					'systemd-nspawn',
					'-D',
					self.path,
					'--timezone=off',
					'-b',
					'--no-pager',
					'--machine',
					self.container_name,
				]
			)

		if not self.ready and self.session:
			while self.session.is_alive():
				if b' login:' in self.session:
					self.ready = True
					break

		Boot._active_boot = self
		return self

	def __exit__(self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None) -> None:
		# b''.join(sys_command('sync')) # No need to, since the underlying fs() object will call sync.
		# TODO: https://stackoverflow.com/questions/28157929/how-to-safely-handle-an-exception-inside-a-context-manager

		if exc_type is not None:
			error(
				str(exc_value),
				f'The error above occurred in a temporary boot-up of the installation {self.path!r}',
			)

		shutdown = None
		shutdown_exit_code: int | None = -1

		try:
			shutdown = SysCommand(f'systemd-run --machine={self.container_name} --pty shutdown now')
		except SysCallError as err:
			shutdown_exit_code = err.exit_code

		if self.session:
			while self.session.is_alive():
				time.sleep(0.25)

		if shutdown and shutdown.exit_code:
			shutdown_exit_code = shutdown.exit_code

		if self.session and (self.session.exit_code == 0 or shutdown_exit_code == 0):
			Boot._active_boot = None
		else:
			session_exit_code = self.session.exit_code if self.session else -1

			raise SysCallError(
				f'Could not shut down temporary boot of {self.path!r}: {session_exit_code}/{shutdown_exit_code}',
				exit_code=next(filter(bool, [session_exit_code, shutdown_exit_code])),
			)

	def __iter__(self) -> Iterator[bytes]:
		if self.session:
			yield from self.session

	def __contains__(self, key: bytes) -> bool:
		if self.session is None:
			return False

		return key in self.session

	def is_alive(self) -> bool:
		if self.session is None:
			return False

		return self.session.is_alive()

	def SysCommand(self, cmd: list[str], *args, **kwargs) -> SysCommand:  # type: ignore[no-untyped-def]
		return SysCommand(['systemd-run', f'--machine={self.container_name}', '--pty', *cmd], *args, **kwargs)

	def SysCommandWorker(self, cmd: list[str], *args, **kwargs) -> SysCommandWorker:  # type: ignore[no-untyped-def]
		return SysCommandWorker(['systemd-run', f'--machine={self.container_name}', '--pty', *cmd], *args, **kwargs)



from dataclasses import dataclass
from typing import Self, TypedDict, override

from archinstall.lib.models.config import SubConfig
from archinstall.lib.translationhandler import tr


class PacmanConfigSerialization(TypedDict):
	parallel_downloads: int
	color: bool


@dataclass
class PacmanConfiguration(SubConfig):
	parallel_downloads: int = 5
	color: bool = True

	@classmethod
	def default(cls) -> Self:
		return cls()

	@override
	def json(self) -> PacmanConfigSerialization:
		return {
			'parallel_downloads': self.parallel_downloads,
			'color': self.color,
		}

	@override
	def summary(self) -> str | None:
		if self.color:
			return tr('Color enabled')
		return None

	def preview(self) -> str:
		color_str = str(self.color)
		output = '{}: {}\n'.format(tr('Parallel Downloads'), self.parallel_downloads)
		output += '{}: {}'.format(tr('Color'), color_str)
		return output

	@classmethod
	def parse_arg(cls, args: PacmanConfigSerialization) -> Self:
		config = cls.default()

		if 'parallel_downloads' in args:
			config.parallel_downloads = int(args['parallel_downloads'])
		if 'color' in args:
			config.color = bool(args['color'])

		return config



import stat
from pathlib import Path
from typing import Self


class LPath(Path):
	@classmethod
	def fs_root(cls) -> Self:
		return cls('/')

	def relative_to_root(self) -> Self:
		return self.relative_to(self.fs_root())

	def add_exec(self) -> None:
		"""Add execute permissions (mirrors `chmod +x`)."""
		mode = self.stat().st_mode
		self.chmod(mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

	def remove_exec(self) -> None:
		"""Remove execute permissions (mirrors `chmod -x`)."""
		mode = self.stat().st_mode
		self.chmod(mode & ~(stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH))



from enum import StrEnum, auto
from typing import Final


class Kernel(StrEnum):
	LINUX = auto()
	LINUX_LTS = 'linux-lts'
	LINUX_ZEN = 'linux-zen'
	LINUX_HARDENED = 'linux-hardened'


DEFAULT_KERNEL: Final = Kernel.LINUX



from pathlib import Path
from typing import Final

from archinstall.lib.linux_path import LPath

ARCHISO_MOUNTPOINT: Final = Path('/run/archiso/airootfs')
MIRRORLIST: Final = LPath('/etc/pacman.d/mirrorlist')
PACMAN_CONF: Final = LPath('/etc/pacman.conf')



from typing import override

from archinstall.default_profiles.profile import DisplayServerType, GreeterType, Profile, ProfileType


class XmonadProfile(Profile):
	def __init__(self) -> None:
		super().__init__(
			'Xmonad',
			ProfileType.WindowMgr,
			support_gfx_driver=True,
			display_server=DisplayServerType.Xorg,
		)

	@property
	@override
	def packages(self) -> list[str]:
		return [
			'xmonad',
			'xmonad-contrib',
			'xmonad-extras',
			'xterm',
			'dmenu',
		]

	@property
	@override
	def default_greeter_type(self) -> GreeterType:
		return GreeterType.Lightdm



from typing import TYPE_CHECKING, override

from archinstall.default_profiles.profile import DisplayServerType, Profile, ProfileType

if TYPE_CHECKING:
	from archinstall.lib.installer import Installer


class XorgProfile(Profile):
	def __init__(
		self,
		name: str = 'Xorg',
		profile_type: ProfileType = ProfileType.Xorg,
	):
		super().__init__(
			name,
			profile_type,
			support_gfx_driver=True,
			display_server=DisplayServerType.Xorg,
		)

	@property
	@override
	def packages(self) -> list[str]:
		return [
			'xorg-server',
			'xorg-xinit',
		]

	@override
	def install(self, install_session: Installer) -> None:
		install_session.add_additional_packages(self.packages)



from enum import Enum

from archinstall.lib.installer import Installer
from archinstall.lib.menu.helpers import Selection
from archinstall.lib.models.users import User
from archinstall.lib.translationhandler import tr
from archinstall.tui.menu_item import MenuItem, MenuItemGroup
from archinstall.tui.result import ResultType


class SeatAccess(Enum):
	seatd = 'seatd'
	polkit = 'polkit'


def provision_seat_access(
	install_session: Installer,
	users: list[User],
	seat_access: str,
) -> None:
	if seat_access == SeatAccess.seatd.value:
		for user in users:
			install_session.arch_chroot(f'usermod -a -G seat {user.username}')


async def select_seat_access(profile_name: str, default: str | None) -> SeatAccess:
	header = tr('{} needs access to your seat').format(profile_name)
	header += f' ({tr("collection of hardware devices i.e. keyboard, mouse")})' + '\n'
	header += tr('Choose an option how to give {} access to your hardware').format(profile_name)

	items = [MenuItem(s.value, value=s) for s in SeatAccess]
	group = MenuItemGroup(items, sort_items=True)

	group.set_default_by_value(default)

	result = await Selection[SeatAccess](
		group,
		header=header,
		allow_skip=False,
	).show()

	if result.type_ == ResultType.Selection:
		return result.get_value()
	else:
		raise ValueError('Unexpected result type from seat access selection')



from typing import TYPE_CHECKING

from archinstall.lib.log import debug
from archinstall.lib.models.application import FontsConfiguration

if TYPE_CHECKING:
	from archinstall.lib.installer import Installer


class FontsApp:
	def install(self, install_session: Installer, fonts_config: FontsConfiguration) -> None:
		packages = [f.value for f in fonts_config.fonts]
		debug(f'Installing fonts: {packages}')
		install_session.add_additional_packages(packages)



import re
from dataclasses import dataclass, field
from enum import Enum
from typing import NotRequired, Self, TypedDict, override

from archinstall.lib.log import debug
from archinstall.lib.models.config import SubConfig
from archinstall.lib.translationhandler import tr


class NicType(Enum):
	ISO = 'iso'
	NM = 'nm'
	NM_IWD = 'nm_iwd'
	IWD = 'iwd'
	MANUAL = 'manual'

	def display_msg(self) -> str:
		match self:
			case NicType.ISO:
				return tr('Copy ISO network configuration to installation')
			case NicType.NM:
				return tr('Use Network Manager (default backend)')
			case NicType.NM_IWD:
				return tr('Use Network Manager (iwd backend)')
			case NicType.IWD:
				return tr('Use standalone iwd')
			case NicType.MANUAL:
				return tr('Manual configuration')


class _NicSerialization(TypedDict):
	iface: str | None
	ip: str | None
	dhcp: bool
	gateway: str | None
	dns: list[str]


@dataclass
class Nic:
	iface: str | None = None
	ip: str | None = None
	dhcp: bool = True
	gateway: str | None = None
	dns: list[str] = field(default_factory=list)

	def table_data(self) -> dict[str, str | bool | list[str]]:
		return {
			'iface': self.iface if self.iface else '',
			'ip': self.ip if self.ip else '',
			'dhcp': self.dhcp,
			'gateway': self.gateway if self.gateway else '',
			'dns': self.dns,
		}

	def json(self) -> _NicSerialization:
		return {
			'iface': self.iface,
			'ip': self.ip,
			'dhcp': self.dhcp,
			'gateway': self.gateway,
			'dns': self.dns,
		}

	@classmethod
	def parse_arg(cls, arg: _NicSerialization) -> Self:
		return cls(
			iface=arg.get('iface', None),
			ip=arg.get('ip', None),
			dhcp=arg.get('dhcp', True),
			gateway=arg.get('gateway', None),
			dns=arg.get('dns', []),
		)

	def as_systemd_config(self) -> str:
		match: list[tuple[str, str]] = []
		network: list[tuple[str, str]] = []

		if self.iface:
			match.append(('Name', self.iface))

		if self.dhcp:
			network.append(('DHCP', 'yes'))
		else:
			if self.ip:
				network.append(('Address', self.ip))
			if self.gateway:
				network.append(('Gateway', self.gateway))
			for dns in self.dns:
				network.append(('DNS', dns))

		config = {'Match': match, 'Network': network}

		config_str = ''
		for top, entries in config.items():
			config_str += f'[{top}]\n'
			config_str += '\n'.join(f'{k}={v}' for k, v in entries)
			config_str += '\n\n'

		return config_str


class _NetworkConfigurationSerialization(TypedDict):
	type: str
	nics: NotRequired[list[_NicSerialization]]


@dataclass
class NetworkConfiguration(SubConfig):
	type: NicType
	nics: list[Nic] = field(default_factory=list)

	@override
	def json(self) -> _NetworkConfigurationSerialization:
		config: _NetworkConfigurationSerialization = {'type': self.type.value}
		if self.nics:
			config['nics'] = [n.json() for n in self.nics]

		return config

	@override
	def summary(self) -> str:
		return self.type.display_msg()

	@classmethod
	def parse_arg(cls, config: _NetworkConfigurationSerialization) -> Self | None:
		nic_type = config.get('type', None)
		if not nic_type:
			return None

		match NicType(nic_type):
			case NicType.ISO:
				return cls(NicType.ISO)
			case NicType.NM:
				return cls(NicType.NM)
			case NicType.NM_IWD:
				return cls(NicType.NM_IWD)
			case NicType.IWD:
				return cls(NicType.IWD)
			case NicType.MANUAL:
				nics_arg = config.get('nics', [])
				if nics_arg:
					nics = [Nic.parse_arg(n) for n in nics_arg]
					return cls(NicType.MANUAL, nics)

		return None


@dataclass
class WifiNetwork:
	bssid: str
	frequency: str
	signal_level: str
	flags: str
	ssid: str

	@override
	def __hash__(self) -> int:
		return hash((self.bssid, self.frequency, self.signal_level, self.flags, self.ssid))

	def table_data(self) -> dict[str, str | int]:
		"""Format WiFi data for table display"""
		return {
			'SSID': self.ssid,
			'Signal': f'{self.signal_level} dBm',
			'Frequency': f'{self.frequency} MHz',
			'Security': self.flags,
			'BSSID': self.bssid,
		}

	@classmethod
	def from_wpa(cls, results: str) -> list[Self]:
		entries = []

		for line in results.splitlines():
			line = line.strip()
			if not line:
				continue

			parts = line.split()
			if len(parts) != 5:
				continue

			wifi = cls(bssid=parts[0], frequency=parts[1], signal_level=parts[2], flags=parts[3], ssid=parts[4])
			entries.append(wifi)

		return entries


@dataclass
class WifiConfiguredNetwork:
	network_id: int
	ssid: str
	bssid: str
	flags: list[str]

	@classmethod
	def from_wpa_cli_output(cls, list_networks: str) -> list[Self]:
		"""
		Example output from 'wpa_cli list_networks'

		Selected interface 'wlan0'
		network id / ssid / bssid / flags
		0	WifiGuest any	[CURRENT]
		1		any [DISABLED]
		2		any [DISABLED]
		"""

		lines = list_networks.strip().splitlines()
		lines = lines[1:]  # remove the header row from the wpa_cli output

		networks = []

		for line in lines:
			line = line.strip()
			parts = line.split('\t')

			if len(parts) < 3:
				continue

			try:
				# flags = cls._extract_flags(parts[3])
				flags: list[str] = []

				networks.append(
					cls(
						network_id=int(parts[0]),
						ssid=parts[1],
						bssid=parts[2],
						flags=flags,
					)
				)
			except ValueError, IndexError:
				debug('Parsing error for network output')

		return networks

	@staticmethod
	def _extract_flags(flag_string: str) -> list[str]:
		pattern = r'\[([^\]]+)\]'

		extracted_values = re.findall(pattern, flag_string)

		return extracted_values



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, NotRequired, Self, TypedDict, override

from archinstall.lib.models.config import SubConfig
from archinstall.lib.models.users import Password, User
from archinstall.lib.translationhandler import tr


class U2FLoginConfigSerialization(TypedDict):
	u2f_login_method: str
	passwordless_sudo: bool


class AuthenticationSerialization(TypedDict):
	u2f_config: NotRequired[U2FLoginConfigSerialization]


class U2FLoginMethod(Enum):
	Passwordless = 'passwordless'
	SecondFactor = 'second_factor'

	def display_value(self) -> str:
		match self:
			case U2FLoginMethod.Passwordless:
				return tr('Passwordless login')
			case U2FLoginMethod.SecondFactor:
				return tr('Second factor login')
			case _:
				raise ValueError(f'Unknown type: {self}')


@dataclass
class U2FLoginConfiguration:
	u2f_login_method: U2FLoginMethod
	passwordless_sudo: bool = False

	def json(self) -> U2FLoginConfigSerialization:
		return {
			'u2f_login_method': self.u2f_login_method.value,
			'passwordless_sudo': self.passwordless_sudo,
		}

	@classmethod
	def parse_arg(cls, args: U2FLoginConfigSerialization) -> Self | None:
		u2f_login_method = args.get('u2f_login_method')

		if not u2f_login_method:
			return None

		u2f_config = cls(u2f_login_method=U2FLoginMethod(u2f_login_method))

		u2f_config.u2f_login_method = U2FLoginMethod(u2f_login_method)

		if passwordless_sudo := args.get('passwordless_sudo') is not None:
			u2f_config.passwordless_sudo = passwordless_sudo

		return u2f_config


@dataclass
class AuthenticationConfiguration(SubConfig):
	root_enc_password: Password | None = None
	users: list[User] = field(default_factory=list)
	u2f_config: U2FLoginConfiguration | None = None

	@classmethod
	def parse_arg(cls, args: dict[str, Any]) -> Self:
		auth_config = cls()

		if (u2f_config := args.get('u2f_config')) is not None:
			auth_config.u2f_config = U2FLoginConfiguration.parse_arg(u2f_config)

		if enc_password := args.get('root_enc_password'):
			auth_config.root_enc_password = Password(enc_password=enc_password)

		return auth_config

	@override
	def json(self) -> AuthenticationSerialization:
		config: AuthenticationSerialization = {}

		if self.u2f_config:
			config['u2f_config'] = self.u2f_config.json()

		return config

	@override
	def summary(self) -> list[str]:
		out: list[str] = []

		if self.root_enc_password:
			out.append(tr('Root password set'))

		if self.users:
			out.append(tr('Configured {} user(s)').format(len(self.users)))

		if self.u2f_config:
			out.append(tr('U2F set up'))

		return out

	def has_superuser(self) -> bool:
		return any(u.sudo for u in self.users)

	def has_regular_user(self) -> bool:
		return len(self.users) > 0



from typing import override

from archinstall.default_profiles.profile import DisplayServerType, GreeterType, Profile, ProfileType
from archinstall.lib.installer import Installer
from archinstall.lib.models.users import User


class BspwmProfile(Profile):
	def __init__(self) -> None:
		super().__init__(
			'Bspwm',
			ProfileType.WindowMgr,
			support_gfx_driver=True,
			display_server=DisplayServerType.Xorg,
		)

	@property
	@override
	def packages(self) -> list[str]:
		return [
			'bspwm',
			'sxhkd',
			'dmenu',
			'xdo',
			'rxvt-unicode',
		]

	@property
	@override
	def default_greeter_type(self) -> GreeterType:
		return GreeterType.Lightdm

	@override
	def provision(self, install_session: Installer, users: list[User]) -> None:
		for user in users:
			install_session.arch_chroot('mkdir -p ~/.config/bspwm ~/.config/sxhkd', run_as=user.username)
			install_session.arch_chroot('cp /usr/share/doc/bspwm/examples/bspwmrc ~/.config/bspwm/', run_as=user.username)
			install_session.arch_chroot('cp /usr/share/doc/bspwm/examples/sxhkdrc ~/.config/sxhkd/', run_as=user.username)
			install_session.arch_chroot('chmod +x ~/.config/bspwm/bspwmrc', run_as=user.username)



import readline

from archinstall.lib.args import USER_CONFIG_FILE, USER_CREDS_FILE, ArchConfig
from archinstall.lib.log import debug
from archinstall.lib.menu.helpers import Confirmation, Selection
from archinstall.lib.menu.util import get_password, prompt_dir
from archinstall.lib.translationhandler import tr
from archinstall.tui.menu_item import MenuItem, MenuItemGroup
from archinstall.tui.result import ResultType


async def confirm_config(config: ArchConfig) -> bool:
	header = f'{tr("The specified configuration will be applied")}. '
	header += tr('Would you like to continue?') + '\n'

	group = MenuItemGroup.yes_no()
	group.set_preview_for_all(lambda x: config.user_config_to_json())

	result = await Confirmation(
		group=group,
		header=header,
		allow_skip=False,
		preset=True,
		preview_location='bottom',
		preview_header=tr('Configuration preview'),
	).show()

	if not result.get_value():
		return False

	return True


async def save_config(config: ArchConfig) -> None:
	def preview(item: MenuItem) -> str | None:
		match item.value:
			case 'user_config':
				serialized = config.user_config_to_json()
				return f'{USER_CONFIG_FILE}\n{serialized}'
			case 'user_creds':
				if maybe_serial := config.user_credentials_to_json():
					return f'{USER_CREDS_FILE}\n{maybe_serial}'
				return tr('No configuration')
			case 'all':
				output = [str(USER_CONFIG_FILE)]
				config.user_credentials_to_json()
				output.append(str(USER_CREDS_FILE))
				return '\n'.join(output)
		return None

	items = [
		MenuItem(
			tr('Save user configuration (including disk layout)'),
			value='user_config',
			preview_action=preview,
		),
		MenuItem(
			tr('Save user credentials'),
			value='user_creds',
			preview_action=preview,
		),
		MenuItem(
			tr('Save all'),
			value='all',
			preview_action=preview,
		),
	]

	group = MenuItemGroup(items)
	result = await Selection[str](
		group,
		allow_skip=True,
		preview_location='right',
	).show()

	match result.type_:
		case ResultType.Skip:
			return
		case ResultType.Selection:
			save_option = result.get_value()
		case _:
			raise ValueError('Unhandled return type')

	readline.set_completer_delims('\t\n=')
	readline.parse_and_bind('tab: complete')

	dest_path = await prompt_dir(
		tr('Enter a directory for the configuration(s) to be saved') + '\n',
		allow_skip=True,
	)

	if not dest_path:
		return

	header = tr('Do you want to save the configuration file(s) to {}?').format(dest_path)

	save_result = await Confirmation(
		header=header,
		allow_skip=False,
		preset=True,
	).show()

	match save_result.type_:
		case ResultType.Selection:
			if not save_result.get_value():
				return
		case _:
			return

	debug(f'Saving configuration files to {dest_path.absolute()}')

	header = tr('Do you want to encrypt the user_credentials.json file?')

	enc_result = await Confirmation(
		header=header,
		allow_skip=False,
		preset=False,
	).show()

	enc_password: str | None = None
	if enc_result.type_ == ResultType.Selection:
		if enc_result.get_value():
			password = await get_password(
				header=tr('Credentials file encryption password'),
				allow_skip=True,
			)

			if password:
				enc_password = password.plaintext

	match save_option:
		case 'user_config':
			config.save_user_config(dest_path)
		case 'user_creds':
			config.save_user_creds(dest_path, password=enc_password)
		case 'all':
			config.save(dest_path, creds=True, password=enc_password)



import logging
import os
import sys
import urllib.error
import urllib.request
from enum import Enum
from pathlib import Path

from archinstall.lib.utils.util import timestamp


class Logger:
	def __init__(self, path: Path | None = None) -> None:
		if path is None:
			path = Path('/var/log/archinstall')

		self._path: Path = path

	@property
	def path(self) -> Path:
		return self._path / 'install.log'

	@path.setter
	def path(self, value: Path) -> None:
		self._path = value

	@property
	def directory(self) -> Path:
		return self._path

	def _check_permissions(self) -> None:
		log_file = self.path

		try:
			self._path.mkdir(exist_ok=True, parents=True)
			log_file.touch(exist_ok=True)

			with log_file.open('a') as f:
				f.write('')
		except PermissionError:
			# Fallback to creating the log file in the current folder
			logger._path = Path('./').absolute()

			warn(f'Not enough permission to place log file at {log_file}, creating it in {logger.path} instead')

	def log(self, level: int, content: str) -> None:
		self._check_permissions()

		with self.path.open('a') as f:
			ts = timestamp()
			level_name = logging.getLevelName(level)
			f.write(f'[{ts}] - {level_name} - {content}\n')

	def get_content(self, max_bytes: int | None = None) -> bytes:
		content = self.path.read_bytes()

		if max_bytes is not None:
			size = self.path.stat().st_size

			if size > max_bytes:
				content = content[-max_bytes:]

		return content


logger = Logger()


def _supports_color() -> bool:
	"""
	Found first reference here:
		https://stackoverflow.com/questions/7445658/how-to-detect-if-the-console-does-support-ansi-escape-codes-in-python
	And re-used this:
		https://github.com/django/django/blob/master/django/core/management/color.py#L12

	Return True if the running system's terminal supports color,
	and False otherwise.
	"""
	supported_platform = sys.platform != 'win32' or 'ANSICON' in os.environ

	# isatty is not always implemented, #6223.
	is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
	return supported_platform and is_a_tty


class Font(Enum):
	bold = '1'
	italic = '3'
	underscore = '4'
	blink = '5'
	reverse = '7'
	conceal = '8'


def _stylize_output(
	text: str,
	fg: str,
	bg: str | None,
	reset: bool,
	font: list[Font] = [],
) -> str:
	"""
	Heavily influenced by:
		https://github.com/django/django/blob/ae8338daf34fd746771e0678081999b656177bae/django/utils/termcolors.py#L13
	Color options here:
		https://askubuntu.com/questions/528928/how-to-do-underline-bold-italic-strikethrough-color-background-and-size-i

	Adds styling to a text given a set of color arguments.
	"""
	colors = {
		'black': '0',
		'red': '1',
		'green': '2',
		'yellow': '3',
		'blue': '4',
		'magenta': '5',
		'cyan': '6',
		'white': '7',
		'teal': '8;5;109',  # Extended 256-bit colors (not always supported)
		'orange': '8;5;208',  # https://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html#256-colors
		'darkorange': '8;5;202',
		'gray': '8;5;246',
		'grey': '8;5;246',
		'darkgray': '8;5;240',
		'lightgray': '8;5;256',
	}

	foreground = {key: f'3{colors[key]}' for key in colors}
	background = {key: f'4{colors[key]}' for key in colors}
	code_list = []

	if text == '' and reset:
		return '\x1b[0m'

	code_list.append(foreground[str(fg)])

	if bg:
		code_list.append(background[str(bg)])

	for o in font:
		code_list.append(o.value)

	ansi = ';'.join(code_list)

	return f'\033[{ansi}m{text}\033[0m'


def journal_log(message: str, level: int = logging.DEBUG) -> None:
	try:
		import systemd.journal  # type: ignore[import-not-found]
	except ModuleNotFoundError:
		return

	log_adapter = logging.getLogger('archinstall')
	log_fmt = logging.Formatter('[%(levelname)s]: %(message)s')
	log_ch = systemd.journal.JournalHandler()
	log_ch.setFormatter(log_fmt)
	log_adapter.addHandler(log_ch)
	log_adapter.setLevel(logging.DEBUG)

	log_adapter.log(level, message)


def info(
	*msgs: str,
	level: int = logging.INFO,
	fg: str = 'white',
	bg: str | None = None,
	reset: bool = False,
	font: list[Font] = [],
) -> None:
	log(*msgs, level=level, fg=fg, bg=bg, reset=reset, font=font)


def debug(
	*msgs: str,
	level: int = logging.DEBUG,
	fg: str = 'white',
	bg: str | None = None,
	reset: bool = False,
	font: list[Font] = [],
) -> None:
	log(*msgs, level=level, fg=fg, bg=bg, reset=reset, font=font)


def error(
	*msgs: str,
	level: int = logging.ERROR,
	fg: str = 'red',
	bg: str | None = None,
	reset: bool = False,
	font: list[Font] = [],
) -> None:
	log(*msgs, level=level, fg=fg, bg=bg, reset=reset, font=font)


def warn(
	*msgs: str,
	level: int = logging.WARNING,
	fg: str = 'yellow',
	bg: str | None = None,
	reset: bool = False,
	font: list[Font] = [],
) -> None:
	log(*msgs, level=level, fg=fg, bg=bg, reset=reset, font=font)


def log(
	*msgs: str,
	level: int = logging.INFO,
	fg: str = 'white',
	bg: str | None = None,
	reset: bool = False,
	font: list[Font] = [],
) -> None:
	text = ' '.join(str(x) for x in msgs)

	logger.log(level, text)

	# Attempt to colorize the output if supported
	# Insert default colors and override with **kwargs
	if _supports_color():
		text = _stylize_output(text, fg, bg, reset, font)

	journal_log(text, level=level)

	if level != logging.DEBUG:
		print(text)


def share_install_log(
	paste_url: str,
	max_bytes: int | None = None,
) -> str | None:
	log_path = logger.path

	if not log_path.exists():
		info(f'Log file not found: {log_path}')
		return None

	content = logger.get_content(max_bytes=max_bytes)

	if len(content) == 0:
		info(f'Log file is empty: {log_path}')
		return None

	try:
		req = urllib.request.Request(paste_url, data=content)
		with urllib.request.urlopen(req) as response:
			url = response.read().decode().strip()
	except urllib.error.URLError as e:
		info(f'Upload failed: {e}')
		return None

	if not url.startswith('http'):
		info(f'Unexpected response from {paste_url}: {url[:200]!r}')
		return None

	return url



import argparse
import json
import os
import stat
import sys
import tempfile
import urllib.error
import urllib.parse
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from enum import Enum, StrEnum
from pathlib import Path
from typing import Any, Self
from urllib.request import Request, urlopen

from pydantic import TypeAdapter
from pydantic.dataclasses import dataclass as p_dataclass

from archinstall.lib.crypt import decrypt, encrypt
from archinstall.lib.log import debug, error, logger, warn
from archinstall.lib.menu.util import get_password
from archinstall.lib.models.application import ApplicationConfiguration, ZramConfiguration
from archinstall.lib.models.authentication import AuthenticationConfiguration
from archinstall.lib.models.bootloader import Bootloader, BootloaderConfiguration
from archinstall.lib.models.config import SubConfig
from archinstall.lib.models.device import DiskEncryption, DiskLayoutConfiguration
from archinstall.lib.models.locale import LocaleConfiguration
from archinstall.lib.models.mirrors import MirrorConfiguration
from archinstall.lib.models.network import NetworkConfiguration
from archinstall.lib.models.package_types import DEFAULT_KERNEL
from archinstall.lib.models.packages import Repository
from archinstall.lib.models.pacman import PacmanConfiguration
from archinstall.lib.models.profile import ProfileConfiguration
from archinstall.lib.models.users import Password, User, UserSerialization
from archinstall.lib.plugins import load_plugin
from archinstall.lib.translationhandler import Language, tr, translation_handler
from archinstall.lib.utils.format import as_key_value_pair
from archinstall.lib.utils.util import is_valid_path
from archinstall.lib.version import get_version
from archinstall.tui.components import tui


class SubCommand(Enum):
	SHARE_LOG = 'share-log'


@p_dataclass
class Arguments:
	config: Path | None = None
	config_url: str | None = None
	creds: Path | None = None
	creds_url: str | None = None
	creds_decryption_key: str | None = None
	silent: bool = False
	dry_run: bool = False
	script: str | None = None
	mountpoint: Path = Path('/mnt')
	skip_ntp: bool = False
	skip_wkd: bool = False
	skip_boot: bool = False
	debug: bool = False
	offline: bool = False
	no_pkg_lookups: bool = False
	plugin: Path | None = None
	plugin_url: str | None = None
	skip_version_check: bool = False
	skip_wifi_check: bool = False
	advanced: bool = False
	verbose: bool = False

	command: SubCommand | None = None


class ArchConfigType(StrEnum):
	VERSION = 'version'
	SCRIPT = 'script'
	LOCALE_CONFIG = 'locale_config'
	ARCHINSTALL_LANGUAGE = 'archinstall_language'
	DISK_CONFIG = 'disk_config'
	PROFILE_CONFIG = 'profile_config'
	MIRROR_CONFIG = 'mirror_config'
	NETWORK_CONFIG = 'network_config'
	BOOTLOADER_CONFIG = 'bootloader_config'
	APP_CONFIG = 'app_config'
	AUTH_CONFIG = 'auth_config'
	SWAP = 'swap'
	USERS = 'users'
	ROOT_ENC_PASSWORD = 'root_enc_password'
	ENCRYPTION_PASSWORD = 'encryption_password'
	HOSTNAME = 'hostname'
	KERNELS = 'kernels'
	NTP = 'ntp'
	TIMEZONE = 'timezone'
	SERVICES = 'services'
	PACKAGES = 'packages'
	PACMAN_CONFIG = 'pacman_config'
	CUSTOM_COMMANDS = 'custom_commands'

	def text(self) -> str:
		match self:
			case ArchConfigType.ARCHINSTALL_LANGUAGE:
				return tr('ArchInstall Language')
			case ArchConfigType.VERSION:
				return tr('Version')
			case ArchConfigType.SCRIPT:
				return tr('Installation Script')
			case ArchConfigType.LOCALE_CONFIG:
				return tr('Locales')
			case ArchConfigType.DISK_CONFIG:
				return tr('Disk configuration')
			case ArchConfigType.PROFILE_CONFIG:
				return tr('Profile')
			case ArchConfigType.MIRROR_CONFIG:
				return tr('Mirrors and repositories')
			case ArchConfigType.NETWORK_CONFIG:
				return tr('Network')
			case ArchConfigType.BOOTLOADER_CONFIG:
				return tr('Bootloader')
			case ArchConfigType.APP_CONFIG:
				return tr('Application')
			case ArchConfigType.AUTH_CONFIG:
				return tr('Authentication')
			case ArchConfigType.SWAP:
				return tr('Swap')
			case ArchConfigType.HOSTNAME:
				return tr('Hostname')
			case ArchConfigType.KERNELS:
				return tr('Kernels')
			case ArchConfigType.NTP:
				return tr('Automatic time sync (NTP)')
			case ArchConfigType.TIMEZONE:
				return tr('Timezone')
			case ArchConfigType.SERVICES:
				return tr('Services')
			case ArchConfigType.PACKAGES:
				return tr('Additional packages')
			case ArchConfigType.PACMAN_CONFIG:
				return tr('Pacman')
			case ArchConfigType.CUSTOM_COMMANDS:
				return tr('Custom commands')
			case ArchConfigType.USERS:
				return tr('Users')
			case ArchConfigType.ROOT_ENC_PASSWORD:
				return tr('Root encrypted password')
			case ArchConfigType.ENCRYPTION_PASSWORD:
				return tr('Disk encryption password')


USER_CONFIG_FILE: Path = Path('user_configuration.json')
USER_CREDS_FILE: Path = Path('user_credentials.json')
DEFAULT_SAVE_PATH = logger.directory


@dataclass
class ArchConfig:
	version: str | None = None
	script: str | None = None
	locale_config: LocaleConfiguration | None = None
	archinstall_language: Language = field(default_factory=lambda: translation_handler.get_language_by_abbr('en'))
	disk_config: DiskLayoutConfiguration | None = None
	profile_config: ProfileConfiguration | None = None
	mirror_config: MirrorConfiguration | None = None
	network_config: NetworkConfiguration | None = None
	bootloader_config: BootloaderConfiguration | None = None
	app_config: ApplicationConfiguration | None = None
	auth_config: AuthenticationConfiguration | None = None
	swap: ZramConfiguration | None = None
	hostname: str = 'archlinux'
	kernels: list[str] = field(default_factory=lambda: [DEFAULT_KERNEL.value])
	ntp: bool = True
	packages: list[str] = field(default_factory=list)
	pacman_config: PacmanConfiguration = field(default_factory=PacmanConfiguration.default)
	timezone: str = 'UTC'
	services: list[str] = field(default_factory=list)
	custom_commands: list[str] = field(default_factory=list)

	def unsafe_config(self) -> dict[ArchConfigType, Any]:
		config: dict[ArchConfigType, list[UserSerialization] | str | None] = {}

		if self.auth_config:
			if self.auth_config.users:
				config[ArchConfigType.USERS] = [user.json() for user in self.auth_config.users]

			if self.auth_config.root_enc_password:
				config[ArchConfigType.ROOT_ENC_PASSWORD] = self.auth_config.root_enc_password.enc_password

		if self.disk_config:
			disk_encryption = self.disk_config.disk_encryption
			if disk_encryption and disk_encryption.encryption_password:
				config[ArchConfigType.ENCRYPTION_PASSWORD] = disk_encryption.encryption_password.plaintext

		return config

	def safe_config(self) -> dict[ArchConfigType, Any]:
		base_config: dict[ArchConfigType, Any] = {
			ArchConfigType.VERSION: self.version,
			ArchConfigType.SCRIPT: self.script,
			ArchConfigType.ARCHINSTALL_LANGUAGE: self.archinstall_language.json(),
		}

		base_config.update(self.plain_cfg())
		sub_config = self.sub_cfg()

		for config_type, value in sub_config.items():
			if not hasattr(value, 'json'):
				raise ValueError(f'Config value for {config_type} must implement json() method')
			base_config[config_type] = value.json()

		return base_config

	def plain_cfg(self) -> dict[ArchConfigType, str | list[str] | bool]:
		return {
			ArchConfigType.HOSTNAME: self.hostname,
			ArchConfigType.KERNELS: self.kernels,
			ArchConfigType.NTP: self.ntp,
			ArchConfigType.TIMEZONE: self.timezone,
			ArchConfigType.SERVICES: self.services,
			ArchConfigType.PACKAGES: self.packages,
			ArchConfigType.CUSTOM_COMMANDS: self.custom_commands,
		}

	def sub_cfg(self) -> dict[ArchConfigType, SubConfig]:
		cfg: dict[ArchConfigType, SubConfig] = {
			ArchConfigType.PACMAN_CONFIG: self.pacman_config,
		}

		if self.mirror_config:
			cfg[ArchConfigType.MIRROR_CONFIG] = self.mirror_config

		if self.bootloader_config:
			cfg[ArchConfigType.BOOTLOADER_CONFIG] = self.bootloader_config

		if self.disk_config:
			cfg[ArchConfigType.DISK_CONFIG] = self.disk_config

		if self.swap:
			cfg[ArchConfigType.SWAP] = self.swap

		if self.auth_config:
			cfg[ArchConfigType.AUTH_CONFIG] = self.auth_config

		if self.locale_config:
			cfg[ArchConfigType.LOCALE_CONFIG] = self.locale_config

		if self.profile_config:
			cfg[ArchConfigType.PROFILE_CONFIG] = self.profile_config

		if self.network_config:
			cfg[ArchConfigType.NETWORK_CONFIG] = self.network_config

		if self.app_config:
			cfg[ArchConfigType.APP_CONFIG] = self.app_config

		return cfg

	@classmethod
	def from_config(cls, args_config: dict[str, Any], args: Arguments) -> Self:
		arch_config = cls()

		arch_config.locale_config = LocaleConfiguration.parse_arg(args_config)

		if script := args_config.get('script', None):
			arch_config.script = script

		if archinstall_lang := args_config.get('archinstall-language', None):
			arch_config.archinstall_language = translation_handler.get_language_by_name(archinstall_lang)
			translation_handler.activate(arch_config.archinstall_language, set_font=False)

		if disk_config := args_config.get('disk_config', {}):
			enc_password = args_config.get('encryption_password', '')
			password = Password(plaintext=enc_password) if enc_password else None
			arch_config.disk_config = DiskLayoutConfiguration.parse_arg(disk_config, password)

			# DEPRECATED
			# backwards compatibility for main level disk_encryption entry
			disk_encryption: DiskEncryption | None = None

			if args_config.get('disk_encryption', None) is not None and arch_config.disk_config is not None:
				disk_encryption = DiskEncryption.parse_arg(
					arch_config.disk_config,
					args_config['disk_encryption'],
					Password(plaintext=args_config.get('encryption_password', '')),
				)

				if disk_encryption:
					arch_config.disk_config.disk_encryption = disk_encryption

		if profile_config := args_config.get('profile_config', None):
			arch_config.profile_config = ProfileConfiguration.parse_arg(profile_config)

		if mirror_config := args_config.get('mirror_config', None):
			backwards_compatible_repo = []
			if additional_repositories := args_config.get('additional-repositories', []):
				backwards_compatible_repo = [Repository(r) for r in additional_repositories]

			arch_config.mirror_config = MirrorConfiguration.parse_args(
				mirror_config,
				backwards_compatible_repo,
			)

		if net_config := args_config.get('network_config', None):
			arch_config.network_config = NetworkConfiguration.parse_arg(net_config)

		if bootloader_config_dict := args_config.get('bootloader_config', None):
			arch_config.bootloader_config = BootloaderConfiguration.parse_arg(bootloader_config_dict, args.skip_boot)
		# DEPRECATED: separate bootloader and uki fields (backward compatibility)
		elif bootloader_str := args_config.get('bootloader', None):
			bootloader = Bootloader.from_arg(bootloader_str, args.skip_boot)
			uki = args_config.get('uki', False)
			if uki and not bootloader.has_uki_support():
				uki = False
			arch_config.bootloader_config = BootloaderConfiguration(bootloader=bootloader, uki=uki, removable=True)

		# deprecated: backwards compatibility
		audio_config_args = args_config.get('audio_config', None)
		app_config_args = args_config.get('app_config', None)

		if audio_config_args is not None or app_config_args is not None:
			arch_config.app_config = ApplicationConfiguration.parse_arg(app_config_args, audio_config_args)

		if auth_config_args := args_config.get('auth_config', None):
			arch_config.auth_config = AuthenticationConfiguration.parse_arg(auth_config_args)

		if hostname := args_config.get('hostname', ''):
			arch_config.hostname = hostname

		if kernels := args_config.get('kernels', []):
			arch_config.kernels = kernels

		arch_config.ntp = args_config.get('ntp', True)

		if packages := args_config.get('packages', []):
			arch_config.packages = packages

		if pacman_config := args_config.get('pacman_config', None):
			arch_config.pacman_config = PacmanConfiguration.parse_arg(pacman_config)
		elif parallel_downloads := args_config.get('parallel_downloads', 0):
			arch_config.pacman_config = PacmanConfiguration(parallel_downloads=int(parallel_downloads))

		swap_arg = args_config.get('swap')
		if swap_arg is not None:
			arch_config.swap = ZramConfiguration.parse_arg(swap_arg)

		if timezone := args_config.get('timezone', 'UTC'):
			arch_config.timezone = timezone

		if services := args_config.get('services', []):
			arch_config.services = services

		# DEPRECATED: backwards compatibility
		root_password = None
		if root_password := args_config.get('!root-password', None):
			root_password = Password(plaintext=root_password)

		if enc_password := args_config.get('root_enc_password', None):
			root_password = Password(enc_password=enc_password)

		if root_password is not None:
			if arch_config.auth_config is None:
				arch_config.auth_config = AuthenticationConfiguration()
			arch_config.auth_config.root_enc_password = root_password

		# DEPRECATED: backwards compatibility
		users: list[User] = []
		if args_users := args_config.get('!users', None):
			users = User.parse_arguments(args_users)

		if args_users := args_config.get('users', None):
			users = User.parse_arguments(args_users)

		if users:
			if arch_config.auth_config is None:
				arch_config.auth_config = AuthenticationConfiguration()
			arch_config.auth_config.users = users

		if custom_commands := args_config.get('custom_commands', []):
			arch_config.custom_commands = custom_commands

		return arch_config

	def user_config_to_json(self) -> str:
		config = self.safe_config()

		adapter = TypeAdapter(dict[ArchConfigType, Any])
		python_dict = adapter.dump_python(config)
		return json.dumps(python_dict, indent=4, sort_keys=True)

	def user_credentials_to_json(self) -> str:
		cfg = self.unsafe_config()

		adapter = TypeAdapter(dict[ArchConfigType, Any])
		python_dict = adapter.dump_python(cfg)
		return json.dumps(python_dict, indent=4, sort_keys=True)

	def write_debug(self) -> None:
		debug(' -- Chosen configuration --')
		debug(self.user_config_to_json())

	def save(
		self,
		dest_path: Path | None = None,
		creds: bool = False,
		password: str | None = None,
	) -> None:
		save_path = dest_path or DEFAULT_SAVE_PATH

		if not is_valid_path(save_path):
			warn(
				f'Destination directory {save_path} does not exist or is not a directory\n.',
				'Configuration files can not be saved',
			)
			return

		self.save_user_config(save_path)
		if creds:
			self.save_user_creds(save_path, password=password)

	def save_user_config(self, dest_path: Path) -> None:
		if not is_valid_path(dest_path):
			error(f'Invalid path {dest_path}. User configuration could not be saved.')
			return

		target = dest_path / USER_CONFIG_FILE
		data = self.user_config_to_json()
		target.write_text(data)
		target.chmod(stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP)

	def save_user_creds(
		self,
		dest_path: Path,
		password: str | None = None,
	) -> None:
		if not is_valid_path(dest_path):
			error(f'Invalid path {dest_path}. User credentials could not be saved.')
			return

		data = self.user_credentials_to_json()

		if password:
			data = encrypt(password, data)

		target = dest_path / USER_CREDS_FILE
		target.write_text(data)
		target.chmod(stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP)

	def as_summary(self) -> str:
		"""
		Render a concise two-column summary of the current configuration.

		Returns an empty string if nothing meaningful to show.
		"""
		cfg: dict[str, str | list[str] | bool] = {}

		for key, value in self.plain_cfg().items():
			cfg[key.text()] = value

		for config_type, obj in self.sub_cfg().items():
			if not hasattr(obj, 'summary'):
				continue

			summary = obj.summary()
			if summary:
				cfg[config_type.text()] = summary

		simple_summary = as_key_value_pair(cfg, ignore_empty=True)

		return simple_summary


class ArchConfigHandler:
	def __init__(self) -> None:
		self._parser: ArgumentParser = self._define_arguments()
		self._add_sub_parsers()

		self._args: Arguments = self._parse_args()
		config = self._parse_config()

		try:
			self._config = ArchConfig.from_config(config, self._args)
			self._config.version = get_version()
		except ValueError as err:
			warn(str(err))
			sys.exit(1)

	@property
	def config(self) -> ArchConfig:
		return self._config

	@property
	def args(self) -> Arguments:
		return self._args

	def get_script(self) -> str:
		if script := self.args.script:
			return script

		if script := self.config.script:
			return script

		return 'guided'

	def print_help(self) -> None:
		self._parser.print_help()

	def _add_sub_parsers(self) -> None:
		subparsers = self._parser.add_subparsers(dest='command', help='Available subcommands')
		_ = subparsers.add_parser(SubCommand.SHARE_LOG.value, help='Upload log file to public server')

	def _define_arguments(self) -> ArgumentParser:
		parser = ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

		parser.add_argument(
			'-v',
			'--version',
			action='version',
			default=False,
			version='%(prog)s ' + get_version(),
		)
		parser.add_argument(
			'--config',
			type=Path,
			nargs='?',
			default=None,
			help='JSON configuration file',
		)
		parser.add_argument(
			'--config-url',
			type=str,
			nargs='?',
			default=None,
			help='Url to a JSON configuration file',
		)
		parser.add_argument(
			'--creds',
			type=Path,
			nargs='?',
			default=None,
			help='JSON credentials configuration file',
		)
		parser.add_argument(
			'--creds-url',
			type=str,
			nargs='?',
			default=None,
			help='Url to a JSON credentials configuration file',
		)
		parser.add_argument(
			'--creds-decryption-key',
			type=str,
			nargs='?',
			default=None,
			help='Decryption key for credentials file',
		)
		parser.add_argument(
			'--silent',
			action='store_true',
			default=False,
			help='WARNING: Disables all prompts for input and confirmation. If no configuration is provided, this is ignored',
		)
		parser.add_argument(
			'--dry-run',
			'--dry_run',
			action='store_true',
			default=False,
			help='Generates a configuration file and then exits instead of performing an installation',
		)
		parser.add_argument(
			'--script',
			nargs='?',
			help='Script to run for installation',
			type=str,
		)
		parser.add_argument(
			'--mountpoint',
			type=Path,
			nargs='?',
			default=Path('/mnt'),
			help='Define an alternate mount point for installation',
		)
		parser.add_argument(
			'--skip-ntp',
			action='store_true',
			help='Disables NTP checks during installation',
			default=False,
		)
		parser.add_argument(
			'--skip-wkd',
			action='store_true',
			help='Disables checking if archlinux keyring wkd sync is complete.',
			default=False,
		)
		parser.add_argument(
			'--skip-boot',
			action='store_true',
			help='Disables installation of a boot loader (note: only use this when problems arise with the boot loader step).',
			default=False,
		)
		parser.add_argument(
			'--debug',
			action='store_true',
			default=False,
			help='Adds debug info into the log',
		)
		parser.add_argument(
			'--offline',
			action='store_true',
			default=False,
			help='Disabled online upstream services such as package search and key-ring auto update.',
		)
		parser.add_argument(
			'--no-pkg-lookups',
			action='store_true',
			default=False,
			help='Disabled package validation specifically prior to starting installation.',
		)
		parser.add_argument(
			'--plugin',
			nargs='?',
			type=Path,
			default=None,
			help='File path to a plugin to load',
		)
		parser.add_argument(
			'--plugin-url',
			type=str,
			nargs='?',
			default=None,
			help='Url to a plugin file to load',
		)
		parser.add_argument(
			'--skip-version-check',
			action='store_true',
			default=False,
			help='Skip the version check when running archinstall',
		)
		parser.add_argument(
			'--skip-wifi-check',
			action='store_true',
			default=False,
			help='Skip wifi check when running archinstall',
		)
		parser.add_argument(
			'--advanced',
			action='store_true',
			default=False,
			help='Enabled advanced options',
		)
		parser.add_argument(
			'--verbose',
			action='store_true',
			default=False,
			help='Enabled verbose options',
		)
		return parser

	def _parse_args(self) -> Arguments:
		argparse_args = vars(self._parser.parse_args())
		args: Arguments = Arguments(**argparse_args)

		# amend the parameters (check internal consistency)
		# Installation can't be silent if config is not passed
		if args.config is None and args.config_url is None:
			args.silent = False

		if args.debug:
			warn(f'Warning: --debug mode will write certain credentials to {logger.path}!')

		if args.plugin:
			load_plugin(args.plugin)

		if args.plugin_url:
			plugin_data = self._fetch_from_url(args.plugin_url)
			plugin_path = self._write_plugin_to_temp_file(plugin_data)
			load_plugin(plugin_path)

		if args.creds_decryption_key is None:
			if os.environ.get('ARCHINSTALL_CREDS_DECRYPTION_KEY'):
				args.creds_decryption_key = os.environ.get('ARCHINSTALL_CREDS_DECRYPTION_KEY')

		return args

	def _parse_config(self) -> dict[str, Any]:
		config: dict[str, Any] = {}
		config_data: str | None = None
		creds_data: str | None = None

		if self._args.config is not None:
			config_data = self._read_file(self._args.config)
		elif self._args.config_url is not None:
			config_data = self._fetch_from_url(self._args.config_url)

		if config_data is not None:
			config.update(json.loads(config_data))

		if self._args.creds is not None:
			creds_data = self._read_file(self._args.creds)
		elif self._args.creds_url is not None:
			creds_data = self._fetch_from_url(self._args.creds_url)

		if creds_data is not None:
			json_data = self._process_creds_data(creds_data)
			if json_data is not None:
				config.update(json_data)

		config = self._cleanup_config(config)

		return config

	def _process_creds_data(self, creds_data: str) -> dict[str, Any] | None:
		if creds_data.startswith('$'):  # encrypted data
			if self._args.creds_decryption_key is not None:
				try:
					creds_data = decrypt(creds_data, self._args.creds_decryption_key)
					return json.loads(creds_data)
				except ValueError as err:
					if 'Invalid password' in str(err):
						error(tr('Incorrect credentials file decryption password'))
						sys.exit(1)
					else:
						debug(f'Error decrypting credentials file: {err}')
						raise err from err
			else:
				header = tr('Enter credentials file decryption password')
				wrong_pwd_text = tr('Incorrect password')
				prompt = header

				while True:
					decryption_pwd: Password | None = tui.run(
						lambda p=prompt: get_password(  # type: ignore[misc]
							header=p,
							allow_skip=False,
							no_confirmation=True,
						)
					)

					if not decryption_pwd:
						return None

					try:
						creds_data = decrypt(creds_data, decryption_pwd.plaintext)
						break
					except ValueError as err:
						if 'Invalid password' in str(err):
							debug('Incorrect credentials file decryption password')
							prompt = f'{header}' + f'\n\n{wrong_pwd_text}'
						else:
							debug(f'Error decrypting credentials file: {err}')
							raise err from err

		return json.loads(creds_data)

	def _fetch_from_url(self, url: str) -> str:
		if urllib.parse.urlparse(url).scheme:
			try:
				req = Request(url, headers={'User-Agent': 'ArchInstall'})
				with urlopen(req) as resp:
					return resp.read().decode('utf-8')
			except urllib.error.HTTPError as err:
				error(f'Could not fetch JSON from {url}: {err}')
		else:
			error('Not a valid url')

		sys.exit(1)

	def _write_plugin_to_temp_file(self, plugin_data: str) -> Path:
		if not plugin_data.strip():
			error('The downloaded plugin is empty')
			sys.exit(1)

		tmp_file = tempfile.NamedTemporaryFile(
			mode='w',
			suffix='.py',
			prefix='archinstall_plugin_',
			delete=False,
		)

		try:
			with tmp_file as f:
				f.write(plugin_data)
		except OSError as err:
			error(f'Could not write the downloaded plugin to a temporary file: {err}')
			sys.exit(1)

		return Path(tmp_file.name)

	def _read_file(self, path: Path) -> str:
		if not path.exists():
			error(f'Could not find file {path}')
			sys.exit(1)

		return path.read_text()

	def _cleanup_config(self, config: Namespace | dict[str, Any]) -> dict[str, Any]:
		clean_args = {}
		for key, val in config.items():
			if isinstance(val, dict):
				val = self._cleanup_config(val)

			if val is not None:
				clean_args[key] = val

		return clean_args



import sys
import time
from pathlib import Path

from archinstall.lib.menu.helpers import Confirmation, Input
from archinstall.lib.models.users import Password, PasswordStrength
from archinstall.lib.translationhandler import tr
from archinstall.tui.components import InputInfo, tui
from archinstall.tui.menu_item import MsgLevelType
from archinstall.tui.result import ResultType


async def get_password(
	header: str | None = None,
	allow_skip: bool = False,
	preset: str | None = None,
	no_confirmation: bool = False,
) -> Password | None:
	def password_hint(value: str) -> InputInfo | None:
		if not value:
			return None
		strength = PasswordStrength.strength(value)
		if strength in (PasswordStrength.VERY_WEAK, PasswordStrength.WEAK):
			return InputInfo(message=tr('Password strength: Weak'), msg_level=MsgLevelType.MsgError)
		elif strength == PasswordStrength.MODERATE:
			return InputInfo(message=tr('Password strength: Moderate'), msg_level=MsgLevelType.MsgWarning)
		elif strength == PasswordStrength.STRONG:
			return InputInfo(message=tr('Password strength: Strong'), msg_level=MsgLevelType.MsgInfo)
		return None

	while True:
		result = await Input(
			header=header,
			allow_skip=allow_skip,
			default_value=preset,
			password=True,
			info_callback=password_hint,
		).show()

		if result.type_ == ResultType.Skip:
			if allow_skip:
				return None
			else:
				continue
		elif result.type_ == ResultType.Selection:
			if not result.get_value():
				if allow_skip:
					return None
				else:
					continue

		password = Password(plaintext=result.get_value())
		break

	if no_confirmation:
		return password

	confirmation_header = f'{tr("Password")}: {password.hidden()}\n\n'
	confirmation_header += tr('Confirm password')

	def _validate(value: str) -> str | None:
		if value != password._plaintext:
			return tr('The password did not match, please try again')
		return None

	result = await Input(
		header=confirmation_header,
		allow_skip=allow_skip,
		password=True,
		validator_callback=_validate,
	).show()

	if result.type_ == ResultType.Skip:
		return None

	return password


async def prompt_dir(
	header: str | None = None,
	validate: bool = True,
	must_exist: bool = True,
	allow_skip: bool = False,
	preset: str | None = None,
) -> Path | None:
	def validate_path(path: str | None) -> str | None:
		if path:
			dest_path = Path(path)

			if must_exist:
				if dest_path.exists() and dest_path.is_dir():
					return None
			else:
				return None

		return tr('Not a valid directory')

	if validate:
		validate_func = validate_path
	else:
		validate_func = None

	result = await Input(
		header=header,
		allow_skip=allow_skip,
		validator_callback=validate_func,
		default_value=preset,
	).show()

	match result.type_:
		case ResultType.Skip:
			return None
		case ResultType.Selection:
			if not result.get_value():
				return None
			return Path(result.get_value())
		case _:
			return None


async def confirm_abort() -> bool:
	prompt = tr('Do you really want to abort?') + '\n'

	result = await Confirmation(
		header=prompt,
		allow_skip=False,
		preset=False,
	).show()

	return result.get_value()


def delayed_warning(message: str) -> bool:
	# Issue a final warning before we continue with something un-revertable.
	# We count down from 5 to 0.
	print(message, end='', flush=True)

	try:
		countdown = '\n5...4...3...2...1\n'
		for c in countdown:
			print(c, end='', flush=True)
			time.sleep(0.25)
	except KeyboardInterrupt:
		ret: bool = tui.run(confirm_abort)
		if ret:
			sys.exit(1)

	return True
